# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is pyramid_whoauth
#
# The Initial Developer of the Original Code is the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2011
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Ryan Kelly (rkelly@mozilla.com)
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
"""

Pyramid authentication policy build on repoze.who.

"""

__ver_major__ = 0
__ver_minor__ = 1
__ver_patch__ = 0
__ver_sub__ = ""
__ver_tuple__ = (__ver_major__, __ver_minor__, __ver_patch__, __ver_sub__)
__version__ = "%d.%d.%d%s" % __ver_tuple__


from repoze.who.api import IAPIFactory

from pyramid.authorization import ACLAuthorizationPolicy

from pyramid_whoauth import utils, auth, views


def includeme(config):
    """Include default whoauth settings into a pyramid config.

    This function provides a hook for pyramid to include the default settings
    for auth via repoze.who.  Activate it like so:

        config.include("pyramid_whoauth")

    It will set up the following defaults for you:

        * add a repoze.who-based AuthenticationPolicy.
        * add a "forbidden view" to invoke repoze.who when auth is required.
        * add default "login" and "logout" routes and views.
        * add a tween to call remember() on response egress.

    """
    # Hook up a default AuthorizationPolicy.
    # ACLAuthorizationPolicy is usually what you want.
    # If the app configures one explicitly then this will get overridden.
    # In auto-commit mode this needs to be set for adding an authn policy.
    authz_policy = ACLAuthorizationPolicy()
    config.set_authorization_policy(authz_policy)

    # Build a WhoAuthenticationPolicy from the deployment settings.
    settings = config.get_settings()
    authn_policy = auth.WhoAuthenticationPolicy.from_settings(settings)
    config.set_authentication_policy(authn_policy)

    # Register the API Factory for other components to find.
    def register_api_factory():
        utils.register_api_factory(config.registry, authn_policy.api_factory)
    config.action(IAPIFactory, register_api_factory)

    # Hook up the challenge view as pyramid's "forbidden view"
    config.add_view(views.challenge, context="pyramid.exceptions.Forbidden")

    # Hook up the login view using configured path and route name.
    login_route = settings.get("who.login_route", "login")
    login_path = settings.get("who.login_path", "/login")
    config.add_route(login_route, login_path)
    config.add_view(views.login, route_name=login_route)

    # Hook up the logout view using configured path and route name.
    logout_route = settings.get("who.logout_route", "logout")
    logout_path = settings.get("who.logout_path", "/logout")
    config.add_route(logout_route, logout_path)
    config.add_view(views.logout, route_name=logout_route)

    # Set up a tween to handle response egress.
    config.add_tween("pyramid_whoauth.tweens.whoauth_tween_factory")



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is pyramid_whoauth
#
# The Initial Developer of the Original Code is the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2011
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Ryan Kelly (rkelly@mozilla.com)
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
"""

IAuthenticationPolicy implementation for pyramid_whoauth

"""

from zope.interface import implements

from pyramid.interfaces import IAuthenticationPolicy
from pyramid.security import Everyone, Authenticated

from repoze.who.utils import resolveDotted

from pyramid_whoauth.utils import (get_api, api_factory_from_settings,
                                   ApplicationRedirectException)


def _null_callback(userid, request):
    """Default group-finder callback for WhoAuthenticationPolicy."""
    return ()


class WhoAuthenticationPolicy(object):
    """Pyramid authentication policy built on top of repoze.who.

    This is a pyramid authentication policy built on top of the repoze.who
    API.  It takes a repoze.who API factory and an optional groupfinder
    callback, and does a straightforward transformation between the repoze.who
    API methods and those of pyramid.
    """

    implements(IAuthenticationPolicy)

    def __init__(self, api_factory=None, callback=None):
        if callback is None:
            callback = _null_callback
        self.api_factory = api_factory
        self._callback = callback

    @classmethod
    def from_settings(cls, settings, prefix="who."):
        """Create a new WhoAuthenticationPolicy from app settings dict."""
        api_factory = api_factory_from_settings(settings, prefix)
        callback = settings.get(prefix + "callback")
        if callback is not None:
            callback = resolveDotted(callback)
            if callback is not None:
                assert callable(callback)
        return cls(api_factory, callback)

    def authenticated_userid(self, request):
        """Get the authenticated userid for the given request.

        This method extracts the userid from the request and passes it
        through two levels of authentication - the repoze.who authenticate
        method, and the configured groupfinder callback.
        """
        userid = self.unauthenticated_userid(request)
        if userid is None:
            return None
        if self._callback(userid, request) is None:
            return None
        return userid

    def unauthenticated_userid(self, request):
        """Get the unauthenticated userid for the given request.

        This method extracts the claimed userid from the request.  Since
        repoze.who does not provide an API to extract the userid without
        authenticating it, the only different between this method and the
        authenticated version is that it does not invoke the groupfinder
        callback function.
        """
        identity = request.environ.get("repoze.who.identity")
        if identity is None:
            api = get_api(request, self.api_factory)
            # Call the repoze.who API to authenticate.
            # If it sets environ["repoze.who.application"] then raise an
            # exception so that this can be taken care of upstream.
            app = request.environ.get("repoze.who.application")
            identity = api.authenticate()
            if app is not request.environ.get("repoze.who.application"):
                raise ApplicationRedirectException
            if identity is None:
                return None
        return identity["repoze.who.userid"]

    def effective_principals(self, request):
        """Get the list of effective principals for the given request.

        This method combines the authenticated userid return by repoze.who
        with the list of groups returned by the groupfinder callback, if any.
        """
        principals = [Everyone]
        userid = self.unauthenticated_userid(request)
        if userid is None:
            return principals
        groups = self._callback(userid, request)
        if groups is None:
            return principals
        principals.insert(0, userid)
        principals.append(Authenticated)
        principals.extend(groups)
        return principals

    def remember(self, request, principal, **kw):
        """Get headers to remember the given principal.

        This method calls the remember() method on all configured repoze.who
        plugins, and returns the combined list of headers.
        """
        identity = {"repoze.who.userid": principal}
        api = get_api(request, self.api_factory)
        #  Give all IIdentifiers a chance to remember the login.
        #  This is the same logic as inside the api.login() method,
        #  but without repeating the authentication step.
        headers = []
        for name, plugin in api.identifiers:
            i_headers = plugin.remember(request.environ, identity)
            if i_headers is not None:
                headers.extend(i_headers)
        return headers

    def forget(self, request):
        """Get headers to forget the identify of the given request.

        This method calls the repoze.who logout() method, which in turn calls
        the forget() method on all configured repoze.who plugins.
        """
        api = get_api(request, self.api_factory)
        return api.logout() or []



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is pyramid_whoauth
#
# The Initial Developer of the Original Code is the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2011
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Ryan Kelly (rkelly@mozilla.com)
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

import unittest2
import tempfile
import base64
from textwrap import dedent

import pyramid.testing
from pyramid.testing import DummyRequest
from pyramid.security import Everyone, Authenticated, authenticated_userid
from pyramid.interfaces import IAuthenticationPolicy, IRequest
from pyramid.response import Response
from pyramid.router import Router
from pyramid.exceptions import Forbidden, NotFound

from zope.interface import implements
from repoze.who.interfaces import IAuthenticator, IIdentifier, IAPIFactory

from pyramid_whoauth.auth import WhoAuthenticationPolicy
from pyramid_whoauth.tweens import whoauth_tween_factory
from pyramid_whoauth.utils import api_factory_from_settings


class DummyAuthenticator(object):
    """Authenticator that accepts login and password."""

    implements(IAuthenticator)

    def authenticate(self, environ, identity):
        if identity.get("login") == identity.get("password"):
            return identity.get("login")


class DummyRememberer(object):
    """Identifier that sets some dummy headers."""

    implements(IIdentifier)

    def identify(self, environ):
        return None

    def remember(self, environ, identity):
        if identity:
            return [("X-Dummy-Remember", "DUMMY")]

    def forget(self, environ, identity):
        return [("X-Dummy-Forget", "DUMMY")]


class DummyRedirector(object):
    """Identifier thar does a dummy redirect from certain URLs."""

    def identify(self, environ):
        if "GO_AWAY" in environ["PATH_INFO"]:
            environ["repoze.who.application"] = self.redirector
        return None

    def remember(self, environ, identity):
        return []

    def forget(self, environ, identity):
        return []

    def redirector(self, environ, start_response):
        start_response("302 Found", [("Location", "/redirected")])
        return ["REDIRECTED"]


def groupfinder(userid, request):
    """Groupfinder that only recognises the "test" user as valid."""
    if userid != "test":
        return None
    return ["group"]


GOOD_AUTHZ = {
  "test": "Basic " + base64.b64encode("test:test"),
  "test2": "Basic " + base64.b64encode("test2:test2")}


BAD_AUTHZ = {
  "test": "Basic " + base64.b64encode("test:badpwd"),
  "test2": "Basic " + base64.b64encode("test2:horseyhorseyneigh")}


SETTINGS = {
    "who.callback": "pyramid_whoauth.tests:groupfinder",
    "who.plugin.basicauth.use": "repoze.who.plugins.basicauth:make_plugin",
    "who.plugin.basicauth.realm": "MyRealm",
    "who.plugin.dummyauth.use": "pyramid_whoauth.tests:DummyAuthenticator",
    "who.plugin.dummyid.use": "pyramid_whoauth.tests:DummyRememberer",
    "who.plugin.dummyredir.use": "pyramid_whoauth.tests:DummyRedirector",
    "who.identifiers.plugins": "dummyid basicauth dummyredir",
    "who.authenticators.plugins": ["dummyauth"],
    "who.challengers.plugins": "basicauth",
    "who.general.challenge_decider":
            "repoze.who.classifiers:default_challenge_decider"}


def raise_forbidden(request):
    """View that always just raises Forbidden."""
    raise Forbidden()


def return_ok(request):
    """View that always just returns "OK" as a string."""
    authenticated_userid(request)
    return Response("OK")


def return_404(request):
    """View that always just returns a 404 response."""
    return Response("Not Found", status="404 Not Found")


def make_request(**kwds):
    environ = {}
    environ["wsgi.version"] = (1, 0)
    environ["wsgi.url_scheme"] = "http"
    environ["SERVER_NAME"] = "localhost"
    environ["SERVER_PORT"] = "80"
    environ["REQUEST_METHOD"] = "GET"
    environ["SCRIPT_NAME"] = ""
    environ["PATH_INFO"] = "/"
    environ.update(kwds)
    return DummyRequest(environ=environ)


class WhoAuthPolicyTests(unittest2.TestCase):

    def setUp(self):
        self.config = pyramid.testing.setUp(autocommit=False)
        self.config.add_settings(**SETTINGS)
        self.config.include("pyramid_whoauth")
        self.config.add_route("forbidden", path="/forbidden")
        self.config.add_view(raise_forbidden, route_name="forbidden")
        self.config.add_route("ok", path="/ok")
        self.config.add_view(return_ok, route_name="ok")
        self.config.add_route("not_found", path="/not_found")
        self.config.add_view(return_404, route_name="not_found")
        self.config.commit()

    def tearDown(self):
        pyramid.testing.tearDown()

    def assertHeadersContain(self, headers, expect_name, expect_value=None):
        for name, value in headers:
            if name == expect_name:
                if expect_value is not None:
                    self.failUnless(expect_value in value)
                break
        else:
            msg = "No %r header was issued"     # pragma: nocover
            assert False, msg % (expect_name,)  # pragma: nocover

    def failIfHeadersContain(self, headers, expect_name, expect_value=None):
        for name, value in headers:
            if name == expect_name:
                if expect_value is None:                     # pragma: nocover
                    msg = "Header %r was present"            # pragma: nocover
                    assert False, msg % (expect_name,)       # pragma: nocover
                elif expect_value in value:                  # pragma: nocover
                    msg = "Header %r contained value %r"     # pragma: nocover
                    msg = msg % (expect_name, expect_value)  # pragma: nocover
                    assert False, msg                        # pragma: nocover

    def test_authenticated_userid(self):
        policy = self.config.registry.getUtility(IAuthenticationPolicy)
        # Authenticated and found by groupfinder.
        req = make_request(HTTP_AUTHORIZATION=GOOD_AUTHZ["test"])
        self.assertEquals(policy.authenticated_userid(req), "test")
        # Authenticated but not found by groupfinder.
        req = make_request(HTTP_AUTHORIZATION=GOOD_AUTHZ["test2"])
        self.assertEquals(policy.authenticated_userid(req), None)
        # Not authenticated
        req = make_request(HTTP_AUTHORIZATION=BAD_AUTHZ["test2"])
        self.assertEquals(policy.authenticated_userid(req), None)

    def test_unauthenticated_userid(self):
        policy = self.config.registry.getUtility(IAuthenticationPolicy)
        # Authenticated and found by groupfinder.
        req = make_request(HTTP_AUTHORIZATION=GOOD_AUTHZ["test"])
        self.assertEquals(policy.unauthenticated_userid(req), "test")
        # Authenticated but not found by groupfinder.
        req = make_request(HTTP_AUTHORIZATION=GOOD_AUTHZ["test2"])
        self.assertEquals(policy.unauthenticated_userid(req), "test2")
        # Not authenticated
        req = make_request(HTTP_AUTHORIZATION=BAD_AUTHZ["test2"])
        self.assertEquals(policy.unauthenticated_userid(req), None)

    def test_effective_principals(self):
        policy = self.config.registry.getUtility(IAuthenticationPolicy)
        # Authenticated and found by groupfinder.
        req = make_request(HTTP_AUTHORIZATION=GOOD_AUTHZ["test"])
        self.assertEquals(sorted(policy.effective_principals(req)),
                          ["group", Authenticated, Everyone, "test"])
        # Authenticated but not found by groupfinder.
        req = make_request(HTTP_AUTHORIZATION=GOOD_AUTHZ["test2"])
        self.assertEquals(sorted(policy.effective_principals(req)),
                          [Everyone])
        # Not authenticated.
        req = make_request(HTTP_AUTHORIZATION=BAD_AUTHZ["test2"])
        self.assertEquals(sorted(policy.effective_principals(req)),
                          [Everyone])

    def test_remember(self):
        policy = self.config.registry.getUtility(IAuthenticationPolicy)
        req = make_request(HTTP_AUTHORIZATION=GOOD_AUTHZ["test"])
        headers = policy.remember(req, "test")
        self.assertEquals(len(headers), 1)
        self.assertEquals(headers[0][0], "X-Dummy-Remember")
        self.assertEquals(headers[0][1], "DUMMY")

    def test_forget(self):
        policy = self.config.registry.getUtility(IAuthenticationPolicy)
        req = make_request(HTTP_AUTHORIZATION=GOOD_AUTHZ["test"])
        headers = policy.forget(req)
        self.assertEquals(len(headers), 2)
        self.assertEquals(headers[0][0], "X-Dummy-Forget")
        self.assertEquals(headers[0][1], "DUMMY")
        self.assertEquals(headers[1][0], "WWW-Authenticate")
        self.assertEquals(headers[1][1], "Basic realm=\"MyRealm\"")

    def test_default_groupfinder(self):
        settings = SETTINGS.copy()
        del settings["who.callback"]
        policy = WhoAuthenticationPolicy.from_settings(settings)
        req = make_request(HTTP_AUTHORIZATION=GOOD_AUTHZ["test"])
        self.assertEquals(sorted(policy.effective_principals(req)),
                          [Authenticated, Everyone, "test"])

    def test_default_api_factory(self):
        policy = WhoAuthenticationPolicy()
        self.assertEquals(policy.api_factory, None)
        req = make_request(HTTP_AUTHORIZATION=GOOD_AUTHZ["test"])
        self.assertEquals(sorted(policy.effective_principals(req)),
                          [Authenticated, Everyone, "test"])

    def test_caching_of_api_factory_from_settings(self):
        policy = self.config.registry.getUtility(IAuthenticationPolicy)
        api_factory = api_factory_from_settings(self.config.registry.settings)
        self.failUnless(api_factory is policy.api_factory)

    def test_settings_from_config_file(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write(dedent("""
            [plugin:basicauth]
            use = repoze.who.plugins.basicauth:make_plugin
            realm = SomeRealm
            [plugin:dummy]
            use = pyramid_whoauth.tests:DummyAuthenticator
            [identifiers]
            plugins = basicauth
            [authenticators]
            plugins = dummy
            [challengers]
            plugins = basicauth
            """))
            f.flush()
            settings = {"who.config_file": f.name}
            policy = WhoAuthenticationPolicy.from_settings(settings)
            req = make_request(HTTP_AUTHORIZATION=GOOD_AUTHZ["test"])
            self.assertEquals(policy.authenticated_userid(req), "test")

    def test_challenge_view_gets_invoked(self):
        app = self.config.make_wsgi_app()
        req = make_request(PATH_INFO="/forbidden")
        def start_response(status, headers):  # NOQA
            self.assertEquals(status, "401 Unauthorized")
            self.assertHeadersContain(headers, "WWW-Authenticate", "MyRealm")
        "".join(app(req.environ, start_response))

    def test_challenge_view_with_no_challengers(self):
        policy = self.config.registry.getUtility(IAuthenticationPolicy)
        del policy.api_factory.challengers[:]
        app = self.config.make_wsgi_app()
        req = make_request(PATH_INFO="/forbidden")
        def start_response(status, headers):  # NOQA
            self.assertEquals(status, "403 Forbidden")
        "".join(app(req.environ, start_response))

    def test_login_view(self):
        app = self.config.make_wsgi_app()

        #  Requesting the login view with no credentials gives a challenge.
        def start_response1(status, headers):
            self.assertEquals(status, "401 Unauthorized")
            self.assertHeadersContain(headers, "WWW-Authenticate", "MyRealm")
        req = make_request(PATH_INFO="/login",
                           QUERY_STRING="came_from=/somewhere")
        "".join(app(req.environ, start_response1))

        #  Requesting the login view with basic-auth creds gives a redirect
        def start_response2(status, headers):
            self.assertEquals(status, "302 Found")
            self.assertHeadersContain(headers, "Location", "/somewhere")
        req = make_request(PATH_INFO="/login",
                           HTTP_AUTHORIZATION=GOOD_AUTHZ["test"],
                           QUERY_STRING="came_from=/somewhere")
        "".join(app(req.environ, start_response2))

        #  Requesting the login view with creds in params gives a redirect
        def start_response3(status, headers):
            self.assertEquals(status, "302 Found")
            self.assertHeadersContain(headers, "Location", "/somewhere")
        query_string = "came_from=/somewhere&login=test&password=test"
        req = make_request(PATH_INFO="/login",
                           QUERY_STRING=query_string)
        "".join(app(req.environ, start_response3))

        #  Requesting the login view with bad creds gives a challenge
        def start_response4(status, headers):
            self.assertEquals(status, "401 Unauthorized")
            self.assertHeadersContain(headers, "WWW-Authenticate", "MyRealm")
        req = make_request(PATH_INFO="/login",
                           HTTP_AUTHORIZATION=BAD_AUTHZ["test"],
                           QUERY_STRING="came_from=/somewhere_outthere")
        "".join(app(req.environ, start_response4))

    def test_logout_view(self):
        app = self.config.make_wsgi_app()

        #  Requesting the logout view with no creds gives challenge+redirect.
        def start_response1(status, headers):
            self.assertEquals(status, "302 Found")
            self.assertHeadersContain(headers, "Location", "/somewhere")
            self.assertHeadersContain(headers, "WWW-Authenticate", "MyRealm")
        req = make_request(PATH_INFO="/logout",
                           QUERY_STRING="came_from=/somewhere")
        "".join(app(req.environ, start_response1))

        #  Requesting the logout view with creds gives challenge+redirect.
        def start_response2(status, headers):
            self.assertEquals(status, "302 Found")
            self.assertHeadersContain(headers, "Location", "/somewhere")
            self.assertHeadersContain(headers, "WWW-Authenticate", "MyRealm")
        req = make_request(PATH_INFO="/logout",
                           HTTP_AUTHORIZATION=GOOD_AUTHZ["test"],
                           QUERY_STRING="came_from=/somewhere")
        "".join(app(req.environ, start_response2))

    def test_tween_sets_remember_headers(self):
        app = self.config.make_wsgi_app()

        #  Requesting a view with no creds should not try to remember me
        def start_response1(status, headers):
            self.assertEquals(status, "200 OK")
            self.failIfHeadersContain(headers, "X-Dummy-Remember")
        req = make_request(PATH_INFO="/ok")
        "".join(app(req.environ, start_response1))

        #  Requesting a view with bad creds should not try to remember me
        def start_response2(status, headers):
            self.assertEquals(status, "200 OK")
            self.failIfHeadersContain(headers, "X-Dummy-Remember")
        req = make_request(PATH_INFO="/ok",
                           HTTP_AUTHORIZATION=BAD_AUTHZ["test"])
        "".join(app(req.environ, start_response2))

        #  Requesting a view with good creds should try to remember me
        def start_response3(status, headers):
            self.assertEquals(status, "200 OK")
            self.assertHeadersContain(headers, "X-Dummy-Remember", "DUMMY")
        req = make_request(PATH_INFO="/ok",
                           HTTP_AUTHORIZATION=GOOD_AUTHZ["test"])
        "".join(app(req.environ, start_response3))

    def test_tween_factory_can_find_api_factory(self):
        registry = self.config.registry
        router = Router(registry)

        # This palaver is necessary to call the tween directly.
        def _make_request(**environ):
            req = make_request(**environ)
            req.__dict__["request_iface"] = IRequest
            req.__dict__["registry"] = registry
            return req

        # Create the factory with no IAPIFactory registered.
        # It should grab the api from the request environ.
        registry.registerUtility(None, IAPIFactory)
        tween = whoauth_tween_factory(router.handle_request, registry)
        req = _make_request(PATH_INFO="/ok",
                            HTTP_AUTHORIZATION=GOOD_AUTHZ["test"])
        response = tween(req)
        self.assertHeadersContain(response.headerlist, "X-Dummy-Remember")

    def test_tween_handles_app_redirects(self):
        app = self.config.make_wsgi_app()

        # Requesting an ordinary not-found view should just get a not-found.
        def start_response1(status, headers):
            self.assertEquals(status, "404 Not Found")
        req = make_request(PATH_INFO="/not_found")
        "".join(app(req.environ, start_response1))

        # Requesting the redirecting view should get the special redirect
        def start_response2(status, headers):
            self.assertEquals(status, "302 Found")
            self.assertHeadersContain(headers, "Location", "redirected")
        req = make_request(PATH_INFO="/GO_AWAY")
        "".join(app(req.environ, start_response2))

        # Remove the Redirector plugin, and we should get the usual NotFound.
        policy = self.config.registry.getUtility(IAuthenticationPolicy)
        policy.api_factory.identifiers.pop()
        req = make_request(PATH_INFO="/GO_AWAY")
        self.assertRaises(NotFound, app, req.environ, start_response1)



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is pyramid_whoauth
#
# The Initial Developer of the Original Code is the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2011
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Ryan Kelly (rkelly@mozilla.com)
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
"""

Tween implementation for pyramid_whoauth.

"""

from pyramid import security
from pyramid.exceptions import NotFound

from pyramid_whoauth.utils import get_api, ApplicationRedirectException


def whoauth_tween_factory(handler, registry):
    """Tween factory for managing repoze.who egress hooks.

    This is a pyramid tween factory that emulates some of the response egress
    hooks from the repoze.who middleware.  In particular:

        * handling plugins that set environ["repoze.who.application"]
        * calling remember() on each authenticated response

    It's useful for compatability with repoze.who plugins that rely on these
    particular behaviours from the middleware, e.g. the OpenID plugin which
    manages its own redirects internally.

    You don't need this if you're actually using the repoze.who middleware
    as a wrapper around your pyramid application.

    This tween does *not* use the challenge decider or issue challenges, since
    pyramid has its own way of doing that.
    """

    def whoauth_tween(request):
        # We have nothing to do on ingress, since the application will call
        # into the repoze.who API as it needs.  Just call the downstream app.
        try:
            # If we're asked to access a non-existent URL, it might be
            # a repoze.who plugin trying to do an internal redirection.
            # to an unknown URL.  Trigger the AuthnPolicy so that it gets
            # a chance to set things up and raise ApplicationRedirectException.
            try:
                response = handler(request)
            except NotFound:
                if "repoze.who.api" not in request.environ:
                    security.unauthenticated_userid(request)
                raise
            else:
                if response.status.startswith("404 "):
                    if "repoze.who.api" not in request.environ:
                        security.unauthenticated_userid(request)
        except ApplicationRedirectException:
            # The AuthnPolicy throws this to indicate that a plugin wants to
            # take control of the response.  Respect any WSGI app that it
            # has put into environ["repoze.who.application"]
            app = request.environ["repoze.who.application"]
            response = request.get_response(app)
        # If there is an identity, make sure it gets remembered.
        # Some plugins depend on this being called on every response rather
        # than explicitly when a new identity is issued.
        identity = request.environ.get("repoze.who.identity")
        if identity:
            api = get_api(request)
            # Give all plugins a chance to remember the login if there is one.
            for name, plugin in api.identifiers:
                i_headers = plugin.remember(request.environ, identity)
                if i_headers is not None:
                    response.headerlist.extend(i_headers)
        return response

    return whoauth_tween



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is pyramid_whoauth
#
# The Initial Developer of the Original Code is the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2011
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Ryan Kelly (rkelly@mozilla.com)
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
"""

Generic view functions for pyramid_whoauth.

"""

from pyramid import security
from pyramid.httpexceptions import HTTPFound, HTTPForbidden
from pyramid.response import Response

from pyramid_whoauth.utils import get_api


def challenge(request, *challenge_args):
    """View that challenges for credentials using repoze.who.

    This method provides a pyramid view that uses the repoze.who challenge
    API to prompt for credentials.  If no challenge can be generated then
    it displays a "403 Forbidden" page.

    You might like to use this as pyramid's "Forbidden View".
    """
    response = None
    api = get_api(request)
    challenge_app = api.challenge(*challenge_args)
    if challenge_app is not None:
        response = request.get_response(challenge_app)
    else:
        response = Response("<h1>Forbidden</h1>", status="403 Forbidden")
    response.headerlist.extend(security.forget(request))
    return response


def login(request):
    """View to process login credentials and remember the user.

    This method provides a pyramid view that uses the repoze.who API
    to authenticate any submitted credentials, then redirects to
    whatever page the user was trying to view.  You can use it as
    a convenient redirection point for plugins that need to submit
    credentials via POST, or as the target for a custom login form.
    """
    came_from = request.params.get("came_from", request.referer or "/")
    # Try to authenticate, either via standard plugin auth
    # or by using the request parameters at the identity.
    userid = security.authenticated_userid(request)
    if userid is not None:
        headers = security.remember(request, userid)
    else:
        api = get_api(request)
        userid, headers = api.login(dict(request.params))
    # If that worked, send them back to where they came from.
    if userid is not None:
        return HTTPFound(location=came_from, headers=headers)
    # If not, trigger the usual forbidden view.
    # In theory this should eventually post back to us again.
    raise HTTPForbidden()


def logout(request):
    """View to forget the logged-in user.

    This method provides a pyramid view that uses the repoze.who API
    to forget any remembered credentials.
    """
    came_from = request.params.get("came_from", request.referer or "/")
    headers = security.forget(request)
    return HTTPFound(location=came_from, headers=headers)



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is pyramid_whoauth
#
# The Initial Developer of the Original Code is the Mozilla Foundation.
# Portions created by the Initial Developer are Copyright (C) 2011
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#   Ryan Kelly (rkelly@mozilla.com)
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
"""

Helper functions for pyramid_whoauth

"""

from repoze.who.config import WhoConfig
from repoze.who.api import IAPIFactory, APIFactory


class ApplicationRedirectException(Exception):
    """Control flow exception for redirecting the downstream app.

    This exception is raised by the WhoAuthenticationPolicy when it detects
    that a plugin is trying to redirect the downstream application by setting
    environ["repoze.who.application"].  It will bubble up as an error in your
    handler unless you have installed the pyramid_whoauth tween, which catches
    the error and does the appropriate redirection.
    """
    pass


def get_api(request, api_factory=None):
    """Get the repoze.who API object for use with the given request.

    This function will lookup and returns the repoze.who API object to use
    for the given request.  If no object exists then it gets the registered
    IAPIFactory utility and uses it to create one.
    """
    # Yes, this is very similar to the logic inside APIFactory itself.
    # I'm re-implementing it so we can look up the default APIFactory
    # in the application registry.
    api = request.environ.get("repoze.who.api")
    if api is None:
        if api_factory is None:
            api_factory = request.registry.getUtility(IAPIFactory)
        api = api_factory(request.environ)
    return api


def register_api_factory(registry, api_factory):
    """Record the APIFactory in the given application registry.

    This function stores a reference to a repoze.who APIFactory in the
    application registry, so that other parts of the application can find
    it at runtime.
    """
    registry.registerUtility(api_factory, IAPIFactory)


def api_factory_from_settings(settings, prefix="who."):
    """Create a new repoze.who APIFactory from the deployment settings.

    This function uses the paster deployment settings to create a repoze.who
    APIFactory object.  Settings starting with "who." are collected and
    formatted into an ini-file which can be read by the repoze.who config
    parser, and the resulting APIFactory object is returned.
    """
    # If we have already built and cached one, just use it directly.
    cache_key = prefix + "api_factory"
    if cache_key in settings:
        return settings[cache_key]

    # Grab out all the settings keys that start with our prefix.
    who_settings = {}
    for name, value in settings.iteritems():
        if not name.startswith(prefix):
            continue
        who_settings[name[len(prefix):]] = value

    # Construct a who.ini config file in memory.
    # First, read in any config file specified in the settings.
    who_ini_lines = []
    if "config_file" in who_settings:
        with open(who_settings["config_file"], "r") as f:
            who_ini_lines.extend(ln.strip() for ln in f)

    # Format any dotted setting names into an ini-file section.
    # For example, a settings file line like:
    #    who.identifiers.plugins = blah
    # Will become an ini-file entry like:
    #    [identifiers]
    #    plugins = blah
    for name, value in who_settings.iteritems():
        if isinstance(value, (list, tuple)):
            value = " ".join(value)
        else:
            value = str(value)
        try:
            section, var = name.rsplit(".", 1)
        except ValueError:
            pass
        else:
            who_ini_lines.append("[%s]" % (section.replace(".", ":"),))
            who_ini_lines.append("%s = %s" % (var, value))

    # Now we can parse that config using repoze.who's own machinery.
    parser = WhoConfig(who_settings.get("here", ""))
    parser.parse("\n".join(who_ini_lines))
    api_factory = APIFactory(parser.identifiers,
                             parser.authenticators,
                             parser.challengers,
                             parser.mdproviders,
                             parser.request_classifier,
                             parser.challenge_decider)

    # Cache it so we don't have to repeat all that work.
    settings[cache_key] = api_factory
    return api_factory
